package fabbroniko.testrawvideofile.h264;

import java.io.InputStream;
import java.io.IOException;

public class InputFileVideoCodecH264 implements CodecH264 {

	/** NAL Maximum length.
	 *
	 */
	public static final int MAX_H264_CODEC_NAL_LENGTH = 65536;

	private byte[] videoBuffer = null;
	private byte[] SPSNALBuffer = null;
	private byte[] PPSNALBuffer = null;
	private int numValidBytes = 0;
	private int numBytesRead = 0;

	private InputStream inputStream = null;
	private boolean isVideoStreamClosed = false;

	public InputFileVideoCodecH264(final InputStream inputStream) {
		videoBuffer = new byte[MAX_H264_CODEC_NAL_LENGTH];
		this.inputStream = inputStream;
	}

	public synchronized byte[] getNextNALUnit(){
		byte[] nalUnit = null;
		int nextNALUnitIndex;

		try {
			numBytesRead = inputStream.read(videoBuffer, numValidBytes, videoBuffer.length - numValidBytes);
		} catch (final IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		// Se il numero di byte letti è diverso da -1 (END OF FILE) allora aggiorna il numero byte di dati validi. 
		if(numBytesRead != -1)
			numValidBytes += numBytesRead;
		// Se il numero di byte validi eccede il numero di byte massimi del buffer, c'è un problema logico di programmazione che deve essere risolto, quindi lancia una eccezione.s
		if(numValidBytes > videoBuffer.length)
			throw new IllegalStateException("Letto un numero di byte maggiore della capacità del buffer.");
		// Se il numero di byte letti è uguale a EOF ed il numero di byte validi nel buffer è 0, allora il flusso dati è terminato, quindi ritorna null. Questo perchè se il flusso di dati è terminato, non è detto che il buffer sia vuoto, quindi gli ultimi byte devono essere consumati.		
		if(numBytesRead == -1 && numValidBytes == 0) {
			isVideoStreamClosed = true;			
			return null;
		}

		nextNALUnitIndex = searchForNextNALUnit(0);

		// Se non è stata trovata nessuna nal unit ritorna null.
 		if(nextNALUnitIndex == -1) {
			isVideoStreamClosed = true;
			return null;
		}
		
		if(nextNALUnitIndex >= numValidBytes)
			throw new IllegalStateException("Indice della prossima NAL Unit superiore al numero di byte validi nel buffer.");

		// IL PROGRAMMA DEVE ARRIVARE FINO A QUI NEI SEGUENTI CASI:
		// - QUANDO IL FLUSSO È CHIUSO, MA NEL BUFFER CI SONO ANCORA DATI.
		// - SEMPRE QUANDO IL FLUSSO È APERTO.
		// --> SE IL FLUSSO È APERTO, MA IL BUFFER NON È PIENO, CHIAMA RICORSIVAMENTE LA FUNZIONE PER GETTARE ALTRI DATI.

		if(nextNALUnitIndex == 0) {
			int nextNALIndex = searchForNextNALUnit(NALUnit.STARTING_SEQUENCE.length);
			if(nextNALIndex == -1)
				nextNALIndex = numValidBytes;

			nalUnit = new byte[nextNALIndex];
			int i;
			for(i = 0; i < nextNALIndex; i++) {
				nalUnit[i] = videoBuffer[i];
			}

			if(isSPSNALUnit(nalUnit)) {
				SPSNALBuffer = nalUnit;
			}else if(isPPSNALUnit(nalUnit)) {
				PPSNALBuffer = nalUnit;
			}

			if(nextNALIndex == MAX_H264_CODEC_NAL_LENGTH) {
				numValidBytes = 0;
			} else {
				shiftBuffer(nextNALIndex, numValidBytes);
			}
		}else{
			shiftBuffer(nextNALUnitIndex, numValidBytes);
		}
		
		return nalUnit;
	}

	private void shiftBuffer(final int begin, final int end) {		
		int i;
		for(i = begin; i < end; i++) {
			videoBuffer[i - begin] = videoBuffer[i];
		}
		numValidBytes = end - begin;
	}

	private int searchForNextNALUnit(final int initialIndex) {
		int i;
		int u;
		for (i = initialIndex, u = 0; i < videoBuffer.length && i < numValidBytes; i++) {
			if(videoBuffer[i] == NALUnit.STARTING_SEQUENCE[u]) {
				u++;
				if(u == NALUnit.STARTING_SEQUENCE.length)
					return i + 1 - NALUnit.STARTING_SEQUENCE.length;
			}else{
				u = 0;
			}
		}
		return -1;
	}

	public static boolean isSPSNALUnit(final byte[] nal) {
		return nal != null && nal.length >= NALUnit.STARTING_SEQUENCE.length + 1 && nal[NALUnit.H264_TYPE_INDEX] == NALUnit.H264_SPS_VALUE;
	}

	public static boolean isPPSNALUnit(final byte[] nal) {
		return nal != null && nal.length >= NALUnit.STARTING_SEQUENCE.length + 1 && nal[NALUnit.H264_TYPE_INDEX] == NALUnit.H264_PPS_VALUE;
	}

	public synchronized boolean isSPSAvailable() {
		return SPSNALBuffer != null;
	}

	public synchronized byte[] getSPSNALUnit() {
		return SPSNALBuffer;
	}

	public synchronized boolean isPPSAvailable() {
		return PPSNALBuffer != null;
	}

	public synchronized byte[] getPPSNALUnit() {
		return PPSNALBuffer;
	}	

	public synchronized boolean isStreamClosed() {
		return isVideoStreamClosed;
	}

	public synchronized void closeStream() {
		try {
			inputStream.close();
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}
}
